package ru.dbotthepony.mc.otm.menu

import com.google.common.collect.ImmutableList
import com.mojang.datafixers.util.Pair
import it.unimi.dsi.fastutil.io.FastByteArrayInputStream
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Reference2ObjectFunction
import it.unimi.dsi.fastutil.objects.Reference2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ReferenceArrayList
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.resources.ResourceLocation
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.Container
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.*
import net.minecraft.world.item.Item
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.enchantment.EnchantmentHelper.hasBindingCurse
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraftforge.network.NetworkEvent
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.compat.cos.cosmeticArmorSlots
import ru.dbotthepony.mc.otm.compat.curios.curiosSlots
import ru.dbotthepony.mc.otm.compat.curios.isCurioSlot
import ru.dbotthepony.mc.otm.container.ItemFilter
import ru.dbotthepony.mc.otm.container.ItemFilterNetworkSlot
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.core.GetterSetter
import ru.dbotthepony.mc.otm.core.util.BigDecimalValueCodec
import ru.dbotthepony.mc.otm.core.util.BinaryStringCodec
import ru.dbotthepony.mc.otm.core.util.BooleanValueCodec
import ru.dbotthepony.mc.otm.core.util.IStreamCodec
import ru.dbotthepony.mc.otm.core.util.ItemValueCodec
import ru.dbotthepony.mc.otm.core.util.NullValueCodec
import ru.dbotthepony.mc.otm.core.util.VarIntValueCodec
import ru.dbotthepony.mc.otm.menu.widget.AbstractWidget
import ru.dbotthepony.mc.otm.network.synchronizer.FieldSynchronizer
import ru.dbotthepony.mc.otm.network.synchronizer.IField
import ru.dbotthepony.mc.otm.network.MatteryPacket
import ru.dbotthepony.mc.otm.network.MatteryPlayerNetworkChannel
import ru.dbotthepony.mc.otm.network.MenuFieldPacket
import ru.dbotthepony.mc.otm.network.MenuNetworkChannel
import ru.dbotthepony.mc.otm.network.SetInventoryFilterPacket
import ru.dbotthepony.mc.otm.network.enqueueWork
import ru.dbotthepony.mc.otm.network.packetHandled
import ru.dbotthepony.mc.otm.network.sender
import java.io.DataInputStream
import java.io.DataOutputStream
import java.math.BigDecimal
import java.util.*
import java.util.function.Predicate
import java.util.function.Supplier
import kotlin.collections.ArrayList

data class PlayerSlot<A : Slot, B : Slot>(val functional: A, val cosmetic: B? = null)

data class EquipmentSlots(
	val armorSlots: List<PlayerSlot<MatteryMenu.EquipmentSlot, Slot>>,
	val curiosSlots: List<PlayerSlot<Slot, Slot>>
)

abstract class MatteryMenu @JvmOverloads protected constructor(
	menuType: MenuType<*>?,
	containerId: Int,
	val inventory: Inventory,
	val tile: BlockEntity? = null
) : AbstractContainerMenu(menuType, containerId) {
	/**
	 * Server->Client synchronizer
	 */
	val mSynchronizer = FieldSynchronizer()
	val ply: Player get() = inventory.player

	private val _matteryWidgets = ArrayList<AbstractWidget>()
	val matteryWidgets: List<AbstractWidget> = Collections.unmodifiableList(_matteryWidgets)

	private val _playerInventorySlots = ArrayList<InventorySlot>()
	private val _playerInventorySlots2 = ArrayList<InventorySlot>()
	private val _playerHotbarSlots = ArrayList<InventorySlot>()
	private val _playerMainSlots = ArrayList<InventorySlot>()
	private val _playerExoSuitSlots = ArrayList<InventorySlot>()
	private val _playerCombinedInventorySlots = ArrayList<InventorySlot>()

	private val playerInputs = ArrayList<PlayerInput<*>>()

	class PlayerInputPacket(val containerId: Int, val inputId: Int, val payload: ByteArray) : MatteryPacket {
		constructor(buff: FriendlyByteBuf) : this(buff.readVarInt(), buff.readVarInt(), ByteArray(buff.readableBytes()).also { buff.readBytes(it) })

		override fun write(buff: FriendlyByteBuf) {
			buff.writeVarInt(containerId)
			buff.writeVarInt(inputId)
			buff.writeBytes(payload)
		}

		override fun play(context: Supplier<NetworkEvent.Context>) {
			context.packetHandled = true

			context.enqueueWork {
				val menu = context.sender?.containerMenu as? MatteryMenu ?: return@enqueueWork
				if (menu.containerId != containerId || !menu.stillValid(context.sender!!)) return@enqueueWork
				val input = menu.playerInputs.getOrNull(inputId) ?: return@enqueueWork
				if (!input.test(context.sender)) return@enqueueWork
				input.invoke(input.codec.read(DataInputStream(FastByteArrayInputStream(payload))))
			}
		}
	}

	/**
	 * Client->Server input
	 */
	inner class PlayerInput<V>(val codec: IStreamCodec<V>, allowSpectators: Boolean = false, val handler: (V) -> Unit) : Predicate<Player?> {
		val id = playerInputs.size
		var allowSpectators by mSynchronizer.bool(allowSpectators)

		init {
			playerInputs.add(this)
		}

		private val filters = ArrayList<Predicate<Player>>()

		init {
			filters.add { allowSpectators || !it.isSpectator }
		}

		fun filter(predicate: Predicate<Player>): PlayerInput<V> {
			filters.add(predicate)
			return this
		}

		override fun test(player: Player?): Boolean {
			if (player == null) return false
			return filters.all { it.test(player) }
		}

		fun input(value: V) {
			if (test(minecraft.player as Player?)) {
				val stream = FastByteArrayOutputStream()
				codec.write(DataOutputStream(stream), value)
				MenuNetworkChannel.sendToServer(PlayerInputPacket(containerId, id, stream.array.copyOfRange(0, stream.length)))
			}
		}

		internal fun invoke(value: Any?) {
			handler.invoke(value as V)
		}
	}

	fun oneWayInput(allowSpectators: Boolean = false, handler: () -> Unit): PlayerInput<Nothing?> {
		return PlayerInput(NullValueCodec, allowSpectators) {
			handler.invoke()
		}
	}

	fun bigDecimalInput(allowSpectators: Boolean = false, handler: (BigDecimal) -> Unit) = PlayerInput(BigDecimalValueCodec, allowSpectators, handler)
	fun booleanInput(allowSpectators: Boolean = false, handler: (Boolean) -> Unit) = PlayerInput(BooleanValueCodec, allowSpectators, handler)
	fun stringInput(allowSpectators: Boolean = false, handler: (String) -> Unit) = PlayerInput(BinaryStringCodec, allowSpectators, handler)
	fun intInput(allowSpectators: Boolean = false, handler: (Int) -> Unit) = PlayerInput(VarIntValueCodec, allowSpectators, handler)

	/**
	 * inventory + exosuit + hotbar (in this order)
	 */
	val playerInventorySlots: List<InventorySlot> = Collections.unmodifiableList(_playerInventorySlots)

	/**
	 * hotbar + inventory + exosuit (in this order)
	 */
	val playerInventorySlots2: List<InventorySlot> = Collections.unmodifiableList(_playerInventorySlots2)

	/**
	 * hotbar only
	 */
	val playerHotbarSlots: List<InventorySlot> = Collections.unmodifiableList(_playerHotbarSlots)

	/**
	 * inventory only
	 */
	val playerMainSlots: List<InventorySlot> = Collections.unmodifiableList(_playerMainSlots)

	/**
	 * exosuit only
	 */
	val playerExoSuitSlots: List<InventorySlot> = Collections.unmodifiableList(_playerExoSuitSlots)

	/**
	 * inventory + exosuit (in this order)
	 */
	val playerCombinedInventorySlots: List<InventorySlot> = Collections.unmodifiableList(_playerCombinedInventorySlots)

	var autoCreateInventoryFrame = true
		private set

	protected val lockedInventorySlots: MutableSet<Int> = HashSet()
	protected open fun isInventorySlotLocked(index: Int): Boolean = lockedInventorySlots.contains(index)

	private val _filterSlots = ArrayList<ItemFilterNetworkSlot>()
	val filterSlots: List<ItemFilterNetworkSlot> = Collections.unmodifiableList(_filterSlots)

	protected var inventorySlotIndexStart = 0
	protected var inventorySlotIndexEnd = 0

	private val playerPacketDistributor = PacketDistributor.PLAYER.with { ply as ServerPlayer }

	fun addFilterSlots(slots: ItemFilter): List<ItemFilterNetworkSlot> {
		val result = ArrayList<ItemFilterNetworkSlot>(slots.size)

		for (i in 0 until slots.size) {
			_filterSlots.add(ItemFilterNetworkSlot(i, _filterSlots.size, slots).also(result::add))
		}

		return result
	}

	fun addFilterSlots(amount: Int): List<ItemFilterNetworkSlot> {
		val result = ArrayList<ItemFilterNetworkSlot>(amount)

		for (i in 0 until amount) {
			_filterSlots.add(ItemFilterNetworkSlot(i, _filterSlots.size, null).also(result::add))
		}

		return result
	}

	fun addWidget(widget: AbstractWidget): Int {
		val indexOf = _matteryWidgets.indexOf(widget)

		if (indexOf == -1) {
			_matteryWidgets.add(widget)
			return _matteryWidgets.size - 1
		}

		return indexOf
	}

	fun getWidget(index: Int): AbstractWidget? {
		if (index !in _matteryWidgets.indices) {
			return null
		}

		return _matteryWidgets[index]
	}

	open inner class InventorySlot(container: Container, index: Int, x: Int = 0, y: Int = 0) : UserFilteredSlot(container, index, x, y) {
		override fun mayPlace(itemStack: ItemStack): Boolean {
			return !isInventorySlotLocked(index) && super.mayPlace(itemStack)
		}

		override fun mayPickup(player: Player): Boolean {
			return !isInventorySlotLocked(index) && super.mayPickup(player)
		}

		override fun isSameInventory(other: Slot): Boolean {
			if (container === inventory || container === ply.matteryPlayer?.exoPackContainer)
				return (other.container === inventory || other.container === ply.matteryPlayer?.exoPackContainer) && isSameFilter(other)

			return super.isSameInventory(other)
		}

		init {
			val mattery = ply.matteryPlayer

			if (mattery != null) {
				if (container === inventory && slotIndex in mattery.regularSlotFilters.indices) {
					filter = GetterSetter.of(
						getter = { mattery.regularSlotFilters[slotIndex].value },
						setter = { MatteryPlayerNetworkChannel.sendToServer(SetInventoryFilterPacket(SetInventoryFilterPacket.Type.INVENTORY, slotIndex, it)) }
					)
				} else if (container === mattery.exoPackContainer) {
					filter = GetterSetter.of(
						getter = { mattery.exoPackContainer.getSlotFilter(slotIndex) },
						setter = { MatteryPlayerNetworkChannel.sendToServer(SetInventoryFilterPacket(SetInventoryFilterPacket.Type.EXOPACK, slotIndex, it)) }
					)
				} else {
					filter = null
				}
			} else {
				filter = null
			}
		}
	}

	open inner class EquipmentSlot(container: Container, index: Int, val type: net.minecraft.world.entity.EquipmentSlot, x: Int = 0, y: Int = 0) : InventorySlot(container, index, x, y) {
		constructor(type: net.minecraft.world.entity.EquipmentSlot, x: Int = 0, y: Int = 0) : this(
			inventory, 34 + type.ordinal, type, x, y
		)

		override fun mayPlace(itemStack: ItemStack): Boolean {
			return super.mayPlace(itemStack) && itemStack.canEquip(type, inventory.player)
		}

		override fun mayPickup(player: Player): Boolean {
			return super.mayPickup(player) && !item.isEmpty && (player.isCreative || !hasBindingCurse(item))
		}

		override fun getMaxStackSize(): Int {
			return 1
		}

		override fun getNoItemIcon(): Pair<ResourceLocation, ResourceLocation>? {
			return Pair.of(InventoryMenu.BLOCK_ATLAS, TEXTURE_EMPTY_SLOTS[type.index])
		}
	}

	protected fun addInventorySlots(autoFrame: Boolean = !ply.isSpectator) {
		check(_playerInventorySlots.isEmpty()) { "Already created inventory slots" }
		autoCreateInventoryFrame = autoFrame

		for (i in 9 .. 35) {
			val slot = InventorySlot(inventory, i)

			_playerInventorySlots.add(slot)
			_playerInventorySlots2.add(slot)
			_playerMainSlots.add(slot)
			_playerCombinedInventorySlots.add(slot)

			mapQuickMove(slot, equipmentSlots)
			mapQuickMoveToExternal(slot)
			addSlot(slot)
		}

		val mattery = ply.matteryPlayer

		if (mattery != null && mattery.hasExoPack) {
			for (i in 0 until mattery.exoPackContainer.containerSize) {
				val slot = InventorySlot(mattery.exoPackContainer, i)

				_playerInventorySlots.add(slot)
				_playerInventorySlots2.add(slot)
				_playerExoSuitSlots.add(slot)
				_playerCombinedInventorySlots.add(slot)

				mapQuickMove(slot, equipmentSlots)
				mapQuickMoveToExternal(slot)
				addSlot(slot)
			}
		}

		for (i in 0..8) {
			val slot = InventorySlot(inventory, i)

			_playerInventorySlots2.add(i, slot)
			_playerInventorySlots.add(slot)
			_playerHotbarSlots.add(slot)

			mapQuickMove(slot, equipmentSlots)
			mapQuickMoveToExternal(slot)
			addSlot(slot)
		}
	}

	override fun broadcastChanges() {
		for (widget in _matteryWidgets) {
			widget.updateServer()
		}

		val payload = mSynchronizer.collectNetworkPayload()

		if (payload != null) {
			MenuNetworkChannel.send(ply, MenuFieldPacket(containerId, payload))
		}

		super.broadcastChanges()

		val consumer = PacketDistributor.PLAYER.with { ply as ServerPlayer }

		for (slot in _filterSlots) {
			val packet = slot.sendChanges()

			if (packet != null) {
				MenuNetworkChannel.send(consumer, packet)
			}
		}
	}

	override fun broadcastFullState() {
		for (widget in _matteryWidgets) {
			widget.updateServer()
		}

		mSynchronizer.invalidate()
		val payload = mSynchronizer.collectNetworkPayload()

		if (payload != null) {
			MenuNetworkChannel.send(ply, MenuFieldPacket(containerId, payload))
		}

		super.broadcastFullState()

		val consumer = PacketDistributor.PLAYER.with { ply as ServerPlayer }

		for (slot in _filterSlots) {
			val packet = slot.sendChanges(true)

			if (packet != null) {
				MenuNetworkChannel.send(consumer, packet)
			}
		}
	}

	override fun stillValid(player: Player): Boolean {
		if (tile == null) return true

		if (player.level.getBlockEntity(tile.blockPos) !== tile) {
			return false
		}

		val pos = tile.blockPos

		return player.distanceToSqr(pos.x.toDouble() + 0.5, pos.y.toDouble() + 0.5, pos.z.toDouble() + 0.5) <= 64.0
	}

	private val externalSlots = ArrayList<Slot>()
	private val quickMoveMapping = Reference2ObjectOpenHashMap<Slot, ReferenceArrayList<Collection<Slot>>>()

	override fun addSlot(pSlot: Slot): Slot {
		if (pSlot in slots) {
			return pSlot
		}

		if (pSlot is UserFilteredSlot && !pSlot.hasSetFilter) {
			val container = pSlot.container

			val input: PlayerInput<Item?>
			val field: IField<Item?>

			if (container is MatteryContainer) {
				input = PlayerInput(ItemValueCodec.nullable, handler = { container.setSlotFilter(pSlot.slotIndex, it) })
				field = mSynchronizer.ComputedField(getter = { container.getSlotFilter(pSlot.slotIndex) }, ItemValueCodec.nullable)
			} else {
				input = PlayerInput(ItemValueCodec.nullable, handler = { throw UnsupportedOperationException() })
				field = mSynchronizer.ComputedField(getter = { null }, ItemValueCodec.nullable)
			}

			pSlot.filter = GetterSetter.of(getter = field::value, setter = input::input)
		}

		return super.addSlot(pSlot)
	}

	/**
	 * Adds slot to "storage slots" list (utilized by quick move) and calls [addSlot]
	 */
	protected fun <T : Slot> addStorageSlot(slot: T, addMapping: Boolean = true): T {
		if (slot !in externalSlots) {
			addSlot(slot)
			externalSlots.add(slot)

			if (addMapping) {
				mapQuickMove(slot, equipmentSlots)
				mapQuickMoveToInventory(slot)
			}
		}

		return slot
	}

	protected fun addStorageSlot(slot: Iterable<Slot>, addMapping: Boolean = true) {
		for (value in slot)
			addStorageSlot(value, addMapping)
	}

	/**
	 * Marks slot as "storage" - shift clicking it will move its contents to Player's inventory
	 */
	protected fun mapQuickMoveToInventory(slot: Slot, prepend: Boolean = false) {
		mapQuickMove(slot, playerInventorySlots, prepend = prepend)
	}

	/**
	 * Marks slot as "inventory" - shift clicking it will move its contents to menu's storage slots
	 */
	protected fun mapQuickMoveToExternal(slot: Slot, prepend: Boolean = false) {
		mapQuickMove(slot, externalSlots, prepend = prepend)
	}

	protected fun mapQuickMove(slot: Slot, target: Collection<Slot>, prepend: Boolean = false) {
		val listing = quickMoveMapping.computeIfAbsent(slot, Reference2ObjectFunction { ReferenceArrayList(1) })
		listing.remove(target)

		if (prepend)
			listing.add(0, target)
		else
			listing.add(target)
	}

	protected fun mapQuickMove(slot: Slot, vararg target: Collection<Slot>) {
		for (value in target) {
			mapQuickMove(slot, value)
		}
	}

	// This method receive Player interactor and slot_index where Shift + Right click occurred
	// It shall return item stack that got moved
	override fun quickMoveStack(ply: Player, slotIndex: Int): ItemStack {
		val slot = slots[slotIndex]

		if (!slot.hasItem() || !slot.mayPickup(ply)) {
			return ItemStack.EMPTY
		}

		val target = quickMoveMapping[slot] ?: return ItemStack.EMPTY
		val copy = slot.item.copy()
		var any = false

		if (target.any { it.any { it is UserFilteredSlot && it.filter != null } }) {
			for (collection in target) {
				if (moveItemStackTo(slot, collection, onlyFiltered = true)) {
					any = true

					if (!slot.hasItem()) {
						return copy
					}
				}
			}
		}

		for (collection in target) {
			if (moveItemStackTo(slot, collection)) {
				any = true

				if (!slot.hasItem()) {
					return copy
				}
			}
		}

		if (any) {
			return copy
		} else {
			return ItemStack.EMPTY
		}
	}

	fun quickMoveToInventory(stack: ItemStack, simulate: Boolean): ItemStack {
		if (_playerInventorySlots.isEmpty()) {
			return stack
		}

		return moveItemStackToSlots(stack, _playerInventorySlots, simulate = simulate)
	}

	override fun canTakeItemForPickAll(itemStack: ItemStack, slot: Slot): Boolean {
		if (slot is EquipmentSlot)
			return false

		return super.canTakeItemForPickAll(itemStack, slot) && (slot !is MatterySlot || slot.canTakeItemForPickAll()) && !slot.isCurioSlot
	}

	override fun moveItemStackTo(
		item: ItemStack,
		initialSlot: Int,
		finalSlot: Int,
		inverse: Boolean
	): Boolean {
		val remainder = moveItemStackToSlots(item, initialSlot, finalSlot, inverse)

		if (remainder.count == item.count) {
			return false
		}

		item.count = remainder.count
		return true
	}

	fun moveItemStackTo(
		source: Slot,
		slots: Collection<Slot>,
		onlyFiltered: Boolean = false
	): Boolean {
		val remainder = moveItemStackToSlots(source.item, slots, onlyFiltered = onlyFiltered)

		if (remainder.count == source.item.count) {
			return false
		}

		if (remainder.isEmpty) {
			source.set(ItemStack.EMPTY)
		} else {
			source.item.count = remainder.count
			source.setChanged()
		}

		return true
	}

	fun moveItemStackToSlots(item: ItemStack, slots: Collection<Slot>, simulate: Boolean = false, onlyFiltered: Boolean = false): ItemStack {
		if (item.isEmpty) {
			return ItemStack.EMPTY
		}

		val copy = item.copy()

		// first pass - stack with existing slots
		if (copy.isStackable) {
			for (slot in slots) {
				if (onlyFiltered && (slot !is UserFilteredSlot || slot.filter == null || slot.filter!!.get() != item.item)) {
					continue
				} else if (!onlyFiltered && slot is UserFilteredSlot && slot.filter != null && slot.filter!!.get() != null && slot.filter!!.get() != item.item) {
					continue
				}

				val limit = slot.getMaxStackSize(copy)

				if (limit > slot.item.count && slot.mayPlace(item) && ItemStack.isSameItemSameTags(slot.item, copy)) {
					val newCount = (slot.item.count + copy.count).coerceAtMost(limit)
					val diff = newCount - slot.item.count
					copy.count -= diff

					if (!simulate) {
						slot.item.count += diff
						slot.setChanged()
					}

					if (copy.isEmpty) {
						return copy
					}
				}
			}
		}

		// second pass - drop stack into first free slot
		for (slot in slots) {
			if (onlyFiltered && (slot !is UserFilteredSlot || slot.filter == null || slot.filter!!.get() != item.item)) {
				continue
			} else if (!onlyFiltered && slot is UserFilteredSlot && slot.filter != null && slot.filter!!.get() != null && slot.filter!!.get() != item.item) {
				continue
			}

			val limit = slot.getMaxStackSize(copy)

			if (!slot.hasItem() && slot.mayPlace(item)) {
				val newCount = copy.count.coerceAtMost(limit)

				if (!simulate) {
					slot.set(copy.copy().also { it.count = newCount })
					slot.setChanged()
				}

				copy.count -= newCount

				if (copy.isEmpty) {
					return copy
				}
			}
		}

		return copy
	}

	fun moveItemStackToSlots(item: ItemStack, initialSlot: Int, finalSlot: Int, inverse: Boolean = false, simulate: Boolean = false): ItemStack {
		if (initialSlot > finalSlot) {
			return item
		}

		require(initialSlot >= 0) { "Invalid initial slot $initialSlot" }
		require(finalSlot < slots.size) { "Final slot $finalSlot is bigger than total size of array of ${slots.size}" }

		val slots = ArrayList<Slot>(finalSlot - initialSlot + 1)
		var filters = false

		for (i in (if (inverse) finalSlot downTo initialSlot else initialSlot .. finalSlot)) {
			val slot = slots[i]
			slots.add(slot)

			if (slot is InventorySlot && slot.filter != null) {
				filters = true
			}
		}

		if (filters) {
			return moveItemStackToSlots(moveItemStackToSlots(item, slots, simulate, onlyFiltered = true), slots, simulate, onlyFiltered = false)
		}

		return moveItemStackToSlots(item, slots, simulate)
	}

	private var armorSlots: ImmutableList<PlayerSlot<EquipmentSlot, Slot>>? = null
	private var curiosSlots: ImmutableList<PlayerSlot<Slot, Slot>>? = null
	private val equipmentSlots = ArrayList<Slot>()

	fun makeArmorSlots(): List<PlayerSlot<EquipmentSlot, Slot>> {
		if (armorSlots != null) {
			return armorSlots!!
		}

		val cosmetic = ply.cosmeticArmorSlots

		return ImmutableList.of(
			PlayerSlot(EquipmentSlot(net.minecraft.world.entity.EquipmentSlot.HEAD), cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.HEAD)),
			PlayerSlot(EquipmentSlot(net.minecraft.world.entity.EquipmentSlot.CHEST), cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.CHEST)),
			PlayerSlot(EquipmentSlot(net.minecraft.world.entity.EquipmentSlot.LEGS), cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.LEGS)),
			PlayerSlot(EquipmentSlot(net.minecraft.world.entity.EquipmentSlot.FEET), cosmetic?.get(net.minecraft.world.entity.EquipmentSlot.FEET)),
		).also {
			armorSlots = it

			for ((a, b) in it) {
				equipmentSlots.add(a)
				addSlot(a)
				if (b != null) addSlot(b)

				mapQuickMoveToInventory(a)
				if (b != null) mapQuickMoveToInventory(b)
			}
		}
	}

	fun makeEquipmentSlots(): EquipmentSlots {
		return EquipmentSlots(
			armorSlots = makeArmorSlots(),
			curiosSlots = curiosSlots ?: ImmutableList.copyOf(ply.curiosSlots).also {
				for ((a, b) in it) {
					equipmentSlots.add(a)
					addSlot(a)
					if (b != null) addSlot(b)

					mapQuickMoveToInventory(a)
					if (b != null) mapQuickMoveToInventory(b)
				}
			}.also { curiosSlots = it }
		)
	}

	companion object {
		val TEXTURE_EMPTY_SLOTS: List<ResourceLocation> = ImmutableList.of(
			InventoryMenu.EMPTY_ARMOR_SLOT_BOOTS,
			InventoryMenu.EMPTY_ARMOR_SLOT_LEGGINGS,
			InventoryMenu.EMPTY_ARMOR_SLOT_CHESTPLATE,
			InventoryMenu.EMPTY_ARMOR_SLOT_HELMET)
	}
}
